import sys
import toml

from pathlib import Path
from string import Template


def _substitute(value, extra_settings):
    """Recursively substitute the variables in a value.

    - If the input is a string, we do a simple substitution.
    - If the input is a dictionary, we recursively check for a string which
      could have a template and perform substitution if we can.
    - If the input is a list, we iterate all the values and if any one is a
      string, we substitute.
    """
    if isinstance(value, str):
        return Template(value).safe_substitute(**extra_settings)
    if isinstance(value, dict):
        for k, v  in value.items():
            value[k] = _substitute(v, extra_settings)
        return value
    if isinstance(value, list):
        return [_substitute(each, extra_settings) for each in value]
    return value


def update_settings(module_name, extra_settings):
    try:
        module = sys.modules[module_name]
    except KeyError:
        __import__(module_name)
        module = sys.modules[module_name]

    for (member, value) in extra_settings.items():
        if member.isupper() and not member.startswith('_'):
            value = _substitute(value, extra_settings)
            setattr(module, member, value)


def load_settings(module_name, settings_paths):
    """Export dictionary variables and properties to module namespace.

    This will export keys that are all upper case and doesn't begin with
    ``_``. These members will be set as attributes on the module
    ``module_name``.

    :param module_name: Name of the module to be used as Django's settings
        module.
    :param settings_paths: A list of file paths to look for settings. The
        first one found is read.
    """
    for path in settings_paths:
        if Path(path).exists():
            settings_path = path
            break
    else:
        print('Error: None of the config files exist.')
        return

    with open(settings_path) as fd:
        try:
            update_settings(
                module_name, toml.load(fd))
        except toml.decoder.TomlDecodeError as error:
            print('Invalid TOML configuration file: {}'.format(settings_path))
            print(error)
            return



from django_settings_toml import load_settings, update_settings


def test_update_settings_basic():
    # test that update settings actually sets the right attributes.
    update_settings('settings', {'KEY': 'Value'})
    # Now, let's import and see if the above value were set.
    import settings
    assert getattr(settings, 'KEY', None) is not None
    assert settings.KEY == 'Value'
    # Let's update one with multiple words joined with `_`.
    update_settings('settings', {'KEY': 'Value', 'KEY_ANOTHER_2': 'Value2'})
    assert getattr(settings, 'KEY_ANOTHER_2', None) is not None
    assert settings.KEY_ANOTHER_2 == 'Value2'


def test_update_settings_private():
    # test that private keys aren't set.
    update_settings('settings', {'_KEY': 'Value'})
    import settings
    assert getattr(settings, '_KEY', None) is None
    # Test that lowercase keys aren't set either.
    update_settings('settings', {'keys': 'value'})
    assert getattr(settings, '_KEY', None) is None


def test_template_substitution():
    # Test that values can have templates but won't raise errors if they can't
    # be substituted with a value.
    update_settings('settings', {'KEY1': 'Value-${KEY2}'})
    import settings
    assert getattr(settings, 'KEY1', None) is not None
    assert settings.KEY1 == 'Value-${KEY2}'
    # Now, test that values are substituted if they exist.
    update_settings('settings', {'KEY1': 'Value-${KEY2}', 'KEY2': '42'})
    assert settings.KEY1 == 'Value-42'


def test_update_settings_boolean(tmpdir):
    # Test that we can set boolean values for settings.
    p = tmpdir.mkdir('django-settings-toml1').join('example.toml')
    p.write("""\
VALUE = false
""")
    load_settings('settings', [str(p)])
    import settings
    assert getattr(settings, 'VALUE', None) is not None
    assert settings.VALUE is False


def test_update_settings_numerals(tmpdir):
    p = tmpdir.mkdir('django-settings-toml2').join('example.toml')
    p.write("""\
INT_VALUE = 100
""")
    load_settings('settings', [str(p)])
    import settings
    assert getattr(settings, 'INT_VALUE', None) is not None
    assert settings.INT_VALUE == 100


def test_update_settings_dictionary(tmpdir):
    # Test that we can load up Dictionaries.
    p = tmpdir.mkdir('django-settings-toml2').join('example.toml')
    p.write("""\
[LOGGING]
version = 1
disable_existing_loggers = false

[LOGGING.filters.require_debug_false]
'file' = 'django.utils.log.RequireDebugFalse'
""")
    load_settings('settings', [str(p)])
    import settings
    assert getattr(settings, 'LOGGING', None) is not None
    assert settings.LOGGING['version'] == 1
    assert settings.LOGGING['disable_existing_loggers'] is False
    assert settings.LOGGING['filters'] == { 'require_debug_false': {
        'file': 'django.utils.log.RequireDebugFalse'
    }}


def test_update_settings_dictionary_templates(tmpdir):
    # Test that we can load up Dictionaries.
    p = tmpdir.mkdir('django-settings-toml3').join('example.toml')
    p.write("""\
BASE_DIR = './tests'

[LOGGING]
version = 1
disable_existing_loggers = false
file = '${BASE_DIR}/example.log'
""")
    load_settings('settings', [str(p)])
    import settings
    assert getattr(settings, 'LOGGING', None) is not None
    assert settings.LOGGING['version'] == 1
    print(settings.LOGGING)
    assert settings.LOGGING['file'] == './tests/example.log'


def test_update_settings_dictionary_templates(tmpdir):
    # Test that we can load up Dictionaries.
    p = tmpdir.mkdir('django-settings-toml4').join('example.toml')
    p.write("""\
BASE_DIR_2 = './tests'

TEMPLATES = [
  'One',
  'Two',
  'Three',
  '${BASE_DIR_2}/Four',
]
""")
    load_settings('settings', [str(p)])
    import settings
    assert getattr(settings, 'TEMPLATES', None) is not None
    assert './tests/Four' in settings.TEMPLATES


def test_update_settings_uses_right_variable(tmpdir):
    # Test that variables defined in toml file overrides the one defined in the
    # settings itself for substition.
    p = tmpdir.mkdir('django-settings-toml5').join('example.toml')
    p.write("""\
BASE_DIR_TESTING = './tests'

VALUE = '${BASE_DIR_TESTING}/something'
""")
    load_settings('settings', [str(p)])
    import settings
    assert getattr(settings, 'VALUE', None) is not None
    assert settings.VALUE == './tests/something'
